import copy
import datetime
import json
import pytest

from base64 import urlsafe_b64encode
from cryptojwt.key_jar import KeyJar
from cryptojwt.jwk.jwk import key_from_jwk_dict
from cryptojwt.tools import keyconv

from idpyoidc.message.oidc import AccessTokenRequest
from idpyoidc.message.oidc import AuthnToken
from idpyoidc.message.oidc import AuthorizationRequest
from idpyoidc.message.oidc import AuthorizationResponse
from idpyoidc.message.oidc import RegistrationRequest
from idpyoidc.server.exception import UnAuthorizedClient

from saml2.authn_context import PASSWORD
from satosa.attribute_mapping import AttributeMapper
from satosa.context import Context
from satosa_oidcop.idpy_oidcop import OidcOpFrontend
from satosa_oidcop.core.storage.mongo import Mongodb
from satosa.internal import AuthenticationInformation
from satosa.internal import InternalData
from tests.users import USERS
from tests.users import OIDC_USERS

from urllib.parse import urlparse, parse_qsl


CLIENT_1_ID = 'jbxedfmfyc'
CLIENT_1_PASSWD = '19cc69b70d0108f630e52f72f7a3bd37ba4e11678ad1a7434e9818e1'
CLIENT_1_RAT = 'z3PCMmC1HZ1QmXeXGOQMJpWQNQynM4xY'
CLIENT_RED_URL = 'https://127.0.0.1:8090/authz_cb/satosa'
CLIENT_1_SESLOGOUT = 'https://127.0.0.1:8090/session_logout/satosa'

# unused
CLIENT_JWK_DICT = {'keys': [{'kty': 'RSA',
   'use': 'sig',
   'kid': 'OWIzN25HNmY0d0VTNWtMeEstTFU3endZbm9ucjhwTVhfLUdwajU1QS1NMA',
   'n': 'ytqcNOfMII7NT1n4AhG8nWBvNuJ7gfBXdOjde-iPIy0OiXj6CeuXZXdaUsWeUElG3yw03IRyGs6Q1sh1_b3kFTjrfMhj7nX50ZWucZJCO0MLtsPqOfjKCiyOJjb4rSDhDrM2PxNJ_eXEuFzUVB5CPCK2GvKzPBZJvtYmGDnaf0CDH2XcWfUeyrFip2zvJ4wrrb4l-hqngPLhAyNaV3QtzbbXJtQTNPlHYghC_prVj18onHsC68fxQg7OfHmPQq9DVZs24rAb6rqxI0PJwSVbUA89gGjuytQAQEFKzR4AR9bfhZBj6H3X5sOcFg8xg-iOBmQxx7vM_5Dxu1sTIvykFQ',
   'e': 'AQAB',
   'd': 'cQeQhHYoKngHdFCYPWbupu5F6doWoZdu08ixKMqzfxErCXSsNfzc1f_EB1zv0qKR5-Z06e6uubshv1vhSuqU_TJDHLt32zZHZf22PrgVSXoZO9Q8XeL_iN28sxRsSeOJI6y97DVuRBfUHjozYU-e7m0U9T0Im9F7c-dVQKhz0_T6JNr_swwatlL4lrIIc-sAWbTZp30ef8QNze48twxtrhf-NcNIft0GF_jLN3PdQrSe7DmEv6BxSHDG2IKOT-KjmoehkKtc0UTP-465GXFKDQZiMjf7nHh7zdqRCJXPDoS9YFpQFGWvH89p5thBACVMpTOgM3t3eXwUODhwTuXB_Q',
   'p': '-7RuaUjObNfd3iKppj5oSbecRpAa_x_vq5zM6Q0qs0PLyJz6wZRGAvg4pmQFwvGskqOBFXcOPWH4WmmvM6YzuPajIoCMLEojQltmBVaoc-IIo9B7Wry6PXrLjmc-2aFOyECYt8QMkVMC7zcswfA1xy-_gNlXCrA40MLSa6MXv5c',
   'q': 'zlDGTW163TMtXvkah1MvDi9w8qfDHZ9L3iQYhZjvooXIyaRCwQP24EwfYdOeXt1PpKdLbI3ZdeWcjAaqpqJCb-X9BXS_H2KNQkMKqhrKYwrlLEyYkGKsgFTB8Kx1-dgzszV2awR4NJImhl4WdhudUbnD9HDhkOouCICYPeYpbzM'},
  {'kty': 'EC',
   'use': 'sig',
   'kid': 'TjNnM3libWFDMjBpeU5WWmZjVkFqdUFCc0dWd3h2eFNGOW1MSk5VYTVUYw',
   'crv': 'P-256',
   'x': 'AWSp7rX1Obb_D7HZhkjAjND721VZaYp5OuJvc0kxSVE',
   'y': 'X1l4LMpq8dI4idQXUPScuarbyz_a0gq0DjHZhdWWiAw',
   'd': 'd2SBLs_LZIxt2U_sdcjTCLLoMYWli_HYkZ0YIDe2SvE'}]
}
# unused
CLIENT_RSA_KEY = key_from_jwk_dict(CLIENT_JWK_DICT['keys'][0])

CLIENT_CONF = {
        'client_id': CLIENT_1_ID,
        'client_name': "ciro",
        'client_salt': '6flfsj0Z',
        'registration_access_token': CLIENT_1_RAT,
        'registration_client_uri': f'https://127.0.0.1:8000/registration_api?client_id={CLIENT_1_ID}',
        'client_id_issued_at': datetime.datetime.utcnow().timestamp(),
        'client_secret': CLIENT_1_PASSWD,
        'client_secret_expires_at': (datetime.datetime.utcnow() + datetime.timedelta(days=365)).timestamp(),
        'application_type': 'web',
        'contacts': ['ops@example.com'],
        'token_endpoint_auth_method': 'client_secret_basic',
        # TODO for jwe and rfc7523
        # 'jwks_uri': 'https://127.0.0.1:8099/static/jwks.json',
        #'jwks': CLIENT_RSA_KEY.serialize(),
        'redirect_uris': [(CLIENT_RED_URL, {})],
        'post_logout_redirect_uris': [(CLIENT_1_SESLOGOUT, None)],
        'response_types': ['code'],
        'grant_types': ['authorization_code'],
        'allowed_scopes': ['openid', 'profile', 'email', 'offline_access']
}

BASE_URL = "https://idpy.oidc.provid.er"
OIDCOP_CONF = {
  "domain": "localhost",
  "server_name": "localhost",
  "base_url": BASE_URL,
  "storage": {
    "class": "satosa_oidcop.core.storage.mongo.Mongodb",
    "kwargs": {
      "url": "mongodb://127.0.0.1:27017/oidcop",
      "connection_params": {
        # "username": "satosa",
        # "password": "thatpassword",
        "connectTimeoutMS": 2000,
        "socketTimeoutMS": 2000,
        "serverSelectionTimeoutMS": 2000
      }
    },
    "db_name": "oidcop",
    "collections": {
      "session": "session_test",
      "client": "client_test"
    }
  },
  "default_target_backend": "spidSaml2",
  "salt_size": 8,
  "op": {
    "server_info": {
      "add_on": {
        "pkce": {
          "function": "idpyoidc.server.oauth2.add_on.pkce.add_support",
          "kwargs": {
            "code_challenge_method": "S256 S384 S512",
            "essential": False
          }
        }
      },
      "authentication": {
        "user": {
          "acr": "urn:oasis:names:tc:SAML:2.0:ac:classes:InternetProtocolPassword",
          "class": "satosa_oidcop.core.user_authn.SatosaAuthnMethod"
        }
      },
      "authz": {
        "class": "idpyoidc.server.authz.AuthzHandling",
        "kwargs": {
          "grant_config": {
            "expires_in": 43200,
            "usage_rules": {
              "access_token": {},
              "authorization_code": {
                "max_usage": 1,
                "supports_minting": [
                  "access_token",
                  "refresh_token",
                  "id_token"
                ]
              },
              "access_token": {},
              "refresh_token": {
                "supports_minting": [
                  "access_token",
                  "refresh_token"
                ]
              }
            }
          }
        }
      },
      "capabilities": {
        "grant_types_supported": [
          "authorization_code",
          "urn:ietf:params:oauth:grant-type:jwt-bearer",
          "refresh_token"
        ],
        "subject_types_supported": [
          "public",
          "pairwise"
        ],
        "scopes_supported": ["openid", "profile", "offline_access"],
        "claims_supported": ["sub", "given_name", "birthdate", "email"],
        "request_object_signing_alg_values_supported": [
          "RS256",
          "RS384",
          "RS512",
          "ES256",
          "ES384",
          "ES512"
        ]
      },
      "claims_interface": {
        "class": "idpyoidc.server.session.claims.ClaimsInterface",
        "kwargs": {}
      },
      "endpoint": {
        "provider_info": {
          "class": "idpyoidc.server.oidc.provider_config.ProviderConfiguration",
          "kwargs": {
            "client_authn_method": None
          },
          "path": ".well-known/openid-configuration"
        },
        "authorization": {
          "class": "idpyoidc.server.oidc.authorization.Authorization",
          "kwargs": {
            "claims_parameter_supported": True,
            "client_authn_method": None,
            "request_object_encryption_alg_values_supported": [
              "RSA-OAEP",
              "RSA-OAEP-256"
            ],
            "request_parameter_supported": True,
            "request_uri_parameter_supported": True,
            "response_modes_supported": [
              "query",
              "fragment",
              "form_post"
            ],
            "response_types_supported": [
              "code"
            ]
          },
          "path": "OIDC/authorization"
        },
        "token": {
          "class": "idpyoidc.server.oidc.token.Token",
          "kwargs": {
            "client_authn_method": [
              "client_secret_post",
              "client_secret_basic",
              "client_secret_jwt",
              "private_key_jwt"
            ]
          },
          "path": "OIDC/token"
        },
        "userinfo": {
          "class": "idpyoidc.server.oidc.userinfo.UserInfo",
          "kwargs": {
            "claim_types_supported": [
              "normal",
              "aggregated",
              "distributed"
            ],
            "userinfo_encryption_alg_values_supported": [
              "RSA-OAEP",
              "RSA-OAEP-256"
            ],
            "userinfo_signing_alg_values_supported": [
              "RS256",
              "RS512",
              "ES256",
              "ES512"
            ]
          },
          "path": "OIDC/userinfo"
        },
        "introspection": {
          "class": "idpyoidc.server.oauth2.introspection.Introspection",
          "kwargs": {
            "client_authn_method": [
              "client_secret_post",
              "client_secret_basic",
              "client_secret_jwt",
              "private_key_jwt"
            ],
            "release": [
              "username"
            ]
          },
          "path": "OIDC/introspection"
        },
        "registration": {
            "class": "idpyoidc.server.oidc.registration.Registration",
            "kwargs": {
              "client_authn_method": None,
              "client_id_generator": {
                "class": "idpyoidc.server.oidc.registration.random_client_id",
                "kwargs": {}
               },
              "client_secret_expiration_time": 432000
            },
            "path": "OIDC/registration"
        },
        "registration_read": {
            "class": "idpyoidc.server.oidc.read_registration.RegistrationRead",
            "kwargs": {
              "client_authn_method": [
               "bearer_header"
               ]},
            "path": "OIDC/registration_read"
        },
      },
      "httpc_params": {
        "verify": False
      },
      "issuer": "https://localhost:10000",
      "keys": {
        "key_defs": [
          {
            "type": "RSA",
            "use": [
              "sig"
            ]
          },
          {
            "crv": "P-256",
            "type": "EC",
            "use": [
              "sig"
            ]
          }
        ],
        "private_path": "data/oidc_op/private/jwks.json",
        "public_path": "data/static/jwks.json",
        "read_only": False,
        "uri_path": "OIDC/jwks.json"
      },
      "login_hint2acrs": {
        "class": "idpyoidc.server.login_hint.LoginHint2Acrs",
        "kwargs": {
          "scheme_map": {
            "email": [
              "urn:oasis:names:tc:SAML:2.0:ac:classes:InternetProtocolPassword"
            ]
          }
        }
      },
      "session_params": {
        "password": "__password_used_to_encrypt_access_token_sid_value",
        "salt": "salt involved in session sub hash ",
        "sub_func": {
          "pairwise": {
            "class": "idpyoidc.server.session.manager.PairWiseID",
            "kwargs": {
              "salt": "CHANGE_ME_OR_LET_IT_BE_RANDOMIC"
            }
          },
          "public": {
            "class": "idpyoidc.server.session.manager.PublicID",
            "kwargs": {
              "salt": "CHANGE_ME_OR_LET_IT_BE_RANDOMIC"
            }
          }
        }
      },
      "template_dir": "templates",
      "token_handler_args": {
        "code": {
          "kwargs": {
            "lifetime": 600
          }
        },
        "id_token": {
          "class": "idpyoidc.server.token.id_token.IDToken",
          "kwargs": {
            "id_token_encryption_alg_values_supported": [
              "RSA-OAEP",
              "RSA-OAEP-256"
            ],
            "id_token_encryption_enc_values_supported": [
              "A128CBC-HS256",
              "A192CBC-HS384",
              "A256CBC-HS512",
              "A128GCM",
              "A192GCM",
              "A256GCM"
            ],
            "id_token_signing_alg_values_supported": [
              "RS256",
              "RS512",
              "ES256",
              "ES512"
            ]
          }
        },
        "refresh": {
          "kwargs": {
            "lifetime": 86400
          }
        },
        "token": {
          "class": "idpyoidc.server.token.jwt_token.JWTToken",
          "kwargs": {
            "lifetime": 3600
          }
        }
      },
      "userinfo": {
        "class": "satosa_oidcop.core.user_info.SatosaOidcUserInfo"
      }
    }
  }
}

INTERNAL_ATTRIBUTES = {
    "attributes": {"mail": {"saml": ["email"], "openid": ["email"]}}
}

CLIENT_AUTHN_REQUEST = {
    'redirect_uri': CLIENT_RED_URL,
    'scope': 'openid profile email offline_access',
    'response_type': 'code',
    'nonce': 'my_nonce',
    'state': 'my_state',
    # TODO
    # 'code_challenge': 'W-Wr0SA2lQgTEKrJm_t-RFzQtaY_-wXxrCp5PnanTe0',
    # 'code_challenge_method': 'S256',
    'client_id': CLIENT_1_ID,
    "prompt": "consent" # required for scope offline_access
}

msg = {
    "application_type": "web",
    "redirect_uris": [
        "https://client.example.org/callback",
        "https://client.example.org/callback2",
    ],
    "client_name": "My Example",
    "client_name#ja-Jpan-JP": "クライアント名",
    "subject_type": "pairwise",
    "token_endpoint_auth_method": "client_secret_basic",
    # "jwks_uri": "https://client.example.org/my_public_keys.jwks",
    "userinfo_encrypted_response_alg": "RSA-OAEP",
    "userinfo_encrypted_response_enc": "A128CBC-HS256",
    "contacts": ["ve7jtb@example.org", "mary@example.org"],
    "request_uris": [
        "https://client.example.org/rf.txt#qpXaRLh_n93TT",
        "https://client.example.org/rf.txt",
    ],
    "post_logout_redirect_uris": [
        "https://rp.example.com/pl?foo=bar",
        "https://rp.example.com/pl",
    ],
}

CLI_REQ = RegistrationRequest(**msg)


class TestOidcOpFrontend(object):

    def create_frontend(self, mongodb_instance, frontend_config=OIDCOP_CONF):
        # monkey patch
        def monkey_mongo(obj):
            obj.client = mongodb_instance

        # will use in-memory storage
        frontend = OidcOpFrontend(lambda ctx, req: None, INTERNAL_ATTRIBUTES,
                                  frontend_config, BASE_URL, "oidc_frontend")
        frontend.register_endpoints(["foo_backend"])
        return frontend

    def clean_inmemory(self, frontend):
        # clean up cdb
        _ec = frontend.app.server.context
        _ec.cdb = {}
        # sman
        frontend._flush_endpoint_context_memory()

    @pytest.fixture
    def frontend(self, mongodb_instance):
        return self.create_frontend(mongodb_instance, OIDCOP_CONF)

    def test_jwks_endpoint(self, context, frontend):
        res = frontend.jwks_endpoint(context)
        assert res._status == "200 OK"
        msg = json.loads(res.message)
        assert msg.get('keys')

    def test_provider_info_endpoint(self, context, frontend):
        res = frontend.provider_info_endpoint(context)
        assert res._status == "200 OK"
        msg = json.loads(res.message)
        assert msg.get('token_endpoint_auth_methods_supported')

    def get_authn_req(self, **kwargs):
        """ produces default or customized oidc autz requests """
        data = copy.deepcopy(CLIENT_AUTHN_REQUEST)
        if kwargs:
            data.update(kwargs)

        req = AuthorizationRequest(**data)
        return req

    @pytest.fixture
    def authn_req(self):
        return self.get_authn_req()

    def insert_client_in_client_db(self, frontend, **kwargs):
        client_conf = copy.deepcopy(CLIENT_CONF)
        client_conf.update(kwargs)

        frontend.app.storage.insert_client(client_conf)
        client = frontend.app.storage.get_client_by_id(
                    client_conf['client_id'])
        return client

    def prepare_call(self, context, frontend, authn_req):
        frontend.auth_req_callback_func = lambda x, y: x
        client = self.insert_client_in_client_db(frontend)
        context.request = authn_req.to_dict()
        return client

    def test_authorization_endpoint(self, context, frontend, authn_req):
        client = self.prepare_call(context, frontend, authn_req)
        assert client['client_id'] == authn_req['client_id']
        res = frontend.authorization_endpoint(context)
        assert isinstance(res, Context)

    def test_handle_authn_request(self, context, frontend, authn_req):
        client = self.prepare_call(context, frontend, authn_req)
        res = frontend.handle_authn_request(context)
        assert isinstance(res, Context)

    def test_handle_authn_request_faulty(self, context, frontend, authn_req):
        client = self.prepare_call(context, frontend, authn_req)

        context.request = {
            'scope': 'email', 'response_type': 'code', 'client_id': CLIENT_1_ID
        }
        res = frontend.handle_authn_request(context)
        assert json.loads(res.message)['error_description'] == "Missing required attribute 'redirect_uri'"

        context.request['redirect_uri'] = CLIENT_RED_URL
        res = frontend.handle_authn_request(context)
        assert json.loads(res.message)['error_description'] == 'openid not in scope'

        context.request['scope'] = 'openid'
        res = frontend.handle_authn_request(context)
        # now's good :)

    def setup_for_authn_response(self, context, frontend, auth_req):
        context.state[frontend.name] = {"oidc_request": auth_req.to_dict()}
        auth_info = AuthenticationInformation(
            PASSWORD, "2021-09-30T12:21:37Z", "unittest_idp.xml"
        )
        internal_response = InternalData(auth_info=auth_info)
        internal_response.attributes = AttributeMapper(
            frontend.internal_attributes
        ).to_internal("saml", USERS["testuser1"])
        internal_response.subject_id = USERS["testuser1"]["eduPersonTargetedID"][0]
        return internal_response

    def test_handle_authn_response_authcode_flow(self, context, frontend, authn_req):
        self.insert_client_in_client_db(frontend, redirect_uri = authn_req["redirect_uri"])
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        http_resp = frontend.handle_authn_response(context, internal_response)
        assert http_resp.message.startswith(authn_req["redirect_uri"])
        assert http_resp.status == '303 See Other'
        _res = urlparse(http_resp.message).query
        resp = AuthorizationResponse().from_urlencoded(_res)

        assert sorted(resp["scope"]) == sorted(authn_req["scope"])
        assert resp["code"]
        assert frontend.name not in context.state
        # Test Token endpoint
        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'client_id': CLIENT_AUTHN_REQUEST['client_id'],
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': resp["code"],
            # TODO
            # 'code_verifier': 'ySfTlMpTEZPYU7H0XQZ75b3B568R5kkMkGRuRpQHOr1KNC9oimGnWygexLJuTyyT'
        }

        credentials = f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        _basic_auth = f"Basic {basic_auth}"
        context.request_authorization = _basic_auth

        # cleanup
        self.clean_inmemory(frontend)

        token_resp = frontend.token_endpoint(context)
        _token_resp = json.loads(token_resp.message)
        assert _token_resp.get('access_token')
        assert _token_resp.get('id_token')

        # cleanup
        self.clean_inmemory(frontend)

        # Test Token endpoint without client ID
        # start new authentication first
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)

        http_resp = frontend.handle_authn_response(context, internal_response)
        _res = urlparse(http_resp.message).query
        resp = AuthorizationResponse().from_urlencoded(_res)
        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': resp["code"],
        }
        token_resp = frontend.token_endpoint(context)
        _token_resp = json.loads(token_resp.message)
        assert _token_resp.get('access_token')
        assert _token_resp.get('id_token')

        # cleanup
        self.clean_inmemory(frontend)

        # Test UserInfo endpoint with FAULTY access_token
        context.request = {}
        _access_token = _token_resp['access_token']
        context.request_authorization = f"{_token_resp['token_type']} {_access_token[:-2]}"
        userinfo_resp = frontend.userinfo_endpoint(context)
        assert userinfo_resp.status == "403 Forbidden"
        
        # Test UserInfo endpoint
        context.request = {}
        _access_token = _token_resp['access_token']
        context.request_authorization = f"{_token_resp['token_type']} {_access_token}"

        userinfo_resp = frontend.userinfo_endpoint(context)
        
        # TODO clientId is not in endpoint_context.cdb and it causes UnknownClient exception. Is it possible it is caused by cleanup?
        _userinfo_resp = json.loads(userinfo_resp.message)
        assert _userinfo_resp.get("sub")

        # cleanup
        self.clean_inmemory(frontend)

        # Test token introspection endpoint
        context.request = {
         'token': _access_token,
         'token_type_hint': 'access_token'
        }
        context.request_authorization = _basic_auth
        introspection_resp = frontend.introspection_endpoint(context)
        assert json.loads(introspection_resp.message).get('sub')

    def test_token_endpoint_without_clientid(self, context, frontend, authn_req):
        self.insert_client_in_client_db(frontend, redirect_uri = authn_req["redirect_uri"])
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        http_resp = frontend.handle_authn_response(context, internal_response)
        #  assert http_resp.message.startswith(authn_req["redirect_uri"])
        #  assert http_resp.status == '303 See Other'
        _res = urlparse(http_resp.message).query
        resp = AuthorizationResponse().from_urlencoded(_res)

        #  assert sorted(resp["scope"]) == sorted(authn_req["scope"])
        #  assert resp["code"]
        #  assert frontend.name not in context.state
        # Test Token endpoint
        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'client_id': CLIENT_AUTHN_REQUEST['client_id'],
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': resp["code"],
        }

        credentials = f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        _basic_auth = f"Basic {basic_auth}"
        context.request_authorization = _basic_auth
    
        # Test Token endpoint without client ID
        # start new authentication first
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        
        http_resp = frontend.handle_authn_response(context, internal_response)
        _res = urlparse(http_resp.message).query
        resp = AuthorizationResponse().from_urlencoded(_res)
        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': resp["code"],
        }
        token_resp = frontend.token_endpoint(context)
        _token_resp = json.loads(token_resp.message)
        assert _token_resp.get('access_token')
        assert _token_resp.get('id_token')

        # cleanup
        self.clean_inmemory(frontend)

    def test_fault_token_endpoint(self, context, frontend):
        # Test Token endpoint
        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'client_id': CLIENT_AUTHN_REQUEST['client_id'],
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': "FAKE-CODE",
            # TODO
            # 'code_verifier': 'ySfTlMpTEZPYU7H0XQZ75b3B568R5kkMkGRuRpQHOr1KNC9oimGnWygexLJuTyyT'
        }
        # and missing credentials ... just to test the exception in satosa
        # for security checks see oidcop tests

        # here fails client auth
        token_resp = frontend.token_endpoint(context)
        assert 'error' in token_resp.message

        # insert the client and miss client auth Basic
        self.insert_client_in_client_db(frontend)
        token_resp = frontend.token_endpoint(context)
        assert 'error' in token_resp.message

        # put also auth basic
        credentials = f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        context.request_authorization = f"Basic {basic_auth}"
        token_resp = frontend.token_endpoint(context)
        assert token_resp.status == '403 Forbidden'
        assert json.loads(token_resp.message)['error'] == 'invalid_request'

    def test_load_cdb_basicauth(self, context, frontend):
        self.insert_client_in_client_db(frontend)
        credentials = f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        context.request_authorization = f"Basic {basic_auth}"
        client = frontend._load_cdb(context)
        assert client

    def test_handle_authn_response_hibrid_flow(self, context, frontend, authn_req):
        response_type = "code id_token token".split(' ')
        self.insert_client_in_client_db(
            frontend,
            response_types = response_type,
            redirect_uri = authn_req["redirect_uri"],
        )
        authn_req['response_type'] = response_type
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        http_resp = frontend.handle_authn_response(context, internal_response)

        res = dict(parse_qsl(urlparse(http_resp.message).query))
        assert res.get('access_token')
        assert res.get('id_token')
        assert res.get('code')

    def test_handle_authn_response_implicit_flow(self, context, frontend, authn_req):
        response_type = "id_token token".split(' ')
        self.insert_client_in_client_db(
            frontend,
            response_types = response_type,
            redirect_uri = authn_req["redirect_uri"],
        )
        authn_req['response_type'] = response_type
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        http_resp = frontend.handle_authn_response(context, internal_response)

        res = dict(parse_qsl(urlparse(http_resp.message).query))
        assert res.get('access_token')
        assert res.get('id_token')

        # tests introspection and userinfo with implicit flow

        # Test UserInfo endpoint
        context.request = {}
        _access_token = res.get('access_token')
        context.request_authorization = f"{res['token_type']} {_access_token}"
        userinfo_resp = frontend.userinfo_endpoint(context)

        _userinfo_resp = json.loads(userinfo_resp.message)
        assert _userinfo_resp.get("sub")

        # Test token introspection endpoint
        credentials = f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        _basic_auth = f"Basic {basic_auth}"

        context.request = {
         'token': _access_token,
         'token_type_hint': 'access_token'
        }
        context.request_authorization = _basic_auth
        introspection_resp = frontend.introspection_endpoint(context)
        assert json.loads(introspection_resp.message).get('sub')

    def test_refresh_token(self, context, frontend, authn_req):
        response_type = "code".split(' ')
        scope = ["openid", "offline_access"]
        self.insert_client_in_client_db(
            frontend,
            response_types = response_type,
            redirect_uri = authn_req["redirect_uri"],
            scope = scope
        )

        authn_req['response_type'] = response_type
        authn_req['scope'] = ["openid", "offline_access"]
        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        http_resp = frontend.handle_authn_response(context, internal_response)
        assert http_resp.message.startswith(authn_req["redirect_uri"])

        assert http_resp.status == '303 See Other'
        _res = urlparse(http_resp.message).query
        resp = AuthorizationResponse().from_urlencoded(_res)

        assert sorted(resp["scope"]) == sorted(authn_req["scope"])
        assert resp["code"]
        assert frontend.name not in context.state

        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'client_id': CLIENT_AUTHN_REQUEST['client_id'],
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': resp["code"],
            # TODO
            # 'code_verifier': 'ySfTlMpTEZPYU7H0XQZ75b3B568R5kkMkGRuRpQHOr1KNC9oimGnWygexLJuTyyT'
        }

        credentials = f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        _basic_auth = f"Basic {basic_auth}"
        context.request_authorization = _basic_auth

        token_resp = frontend.token_endpoint(context)
        res = json.loads(token_resp.message)
        assert res.get('refresh_token')

        # test refresh token
        context.request = {
            "grant_type" : "refresh_token",
            "client_id" : CLIENT_1_ID,
            "client_secret" : CLIENT_1_PASSWD,
            "refresh_token" : res.get('refresh_token')
        }
        refresh_resp = frontend.token_endpoint(context)
        _res = json.loads(refresh_resp.message)
        assert _res.get('refresh_token')
        assert _res.get('access_token')

        # Test FAULTY refresh_token
        context.request["refresh_token"] = _res.get('refresh_token')[:-2]
        refresh_resp = frontend.token_endpoint(context)
        assert refresh_resp.status == "403 Forbidden"
        _res = json.loads(refresh_resp.message)
        assert _res['error'] == "invalid_grant"
        assert _res['error_description'] == "Invalid refresh token"

    # def test_private_key_jwt_token_endpoint(self, context, frontend):
        # """
        # Unused because we cannot have private_key_jwt/RFC7523 in satosa
        # client/rp is a machine, while auth code flow needs an human user

        # in satosa the user MUST call the authorization endpoint
        # """
        # client_assertion_data = {
            # "iss": CLIENT_1_ID,
            # "sub": CLIENT_1_ID,
            # "aud": [f"{BASE_URL}/OIDC/token"],
            # "jti": "my-opaque-jti-value",
            # "exp": (datetime.datetime.now() + datetime.timedelta(days=1)).timestamp(),
            # "iat": datetime.datetime.now().timestamp()

        # }
        # _private_key_jwt = AuthnToken(**client_assertion_data)
        # client_assertion = _private_key_jwt.to_jwt(key=[CLIENT_RSA_KEY], algorithm='RS256')

        # ## Test Token endpoint
        # context.request = {
            # 'grant_type': 'authorization_code',
            # 'redirect_uri': CLIENT_RED_URL,
            # 'state': CLIENT_AUTHN_REQUEST['state'],
            # 'code': "FAKE-CODE",
            # "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            # "client_assertion": client_assertion
        # }
        # self.insert_client_in_client_db(frontend)
        # token_resp = frontend.token_endpoint(context)
        # breakpoint()

    def test_client_registration_endpoint(self, context, frontend, authn_req):
        # just to test reserved client_id
        self.insert_client_in_client_db(frontend)
        context.request = CLI_REQ.to_dict()
        http_resp = frontend.registration_endpoint(context)
        _resp = json.loads(http_resp.message)
        assert _resp['client_id']

    def test_client_registration_read_endpoint(self, context, frontend, authn_req):
        self.insert_client_in_client_db(frontend)
        _bearer_auth = f"Bearer {CLIENT_1_RAT}"
        context.request_authorization = _bearer_auth
        context.request = {'client_id': CLIENT_1_ID}
        http_resp = frontend.registration_read_endpoint(context)
        _resp = json.loads(http_resp.message)
        assert _resp['client_id'] == CLIENT_1_ID
        assert _resp['client_secret'] == CLIENT_1_PASSWD


    def test_handle_authn_authcode_steal_two_clients(self, context, frontend, authn_req):
        # client 1
        self.insert_client_in_client_db(frontend, redirect_uri = authn_req["redirect_uri"])

        # client 2
        self.insert_client_in_client_db(
            frontend,
            client_id="client_2",
            redirect_uri = authn_req["redirect_uri"]
        )

        internal_response = self.setup_for_authn_response(context, frontend, authn_req)
        http_resp = frontend.handle_authn_response(context, internal_response)
        assert http_resp.message.startswith(authn_req["redirect_uri"])
        assert http_resp.status == '303 See Other'
        _res = urlparse(http_resp.message).query
        resp = AuthorizationResponse().from_urlencoded(_res)

        assert sorted(resp["scope"]) == sorted(authn_req["scope"])
        assert resp["code"]
        assert frontend.name not in context.state


        # imagine that client2 steals the auth code and uses it to Token endpoint
        context.request = {
            'grant_type': 'authorization_code',
            'redirect_uri': CLIENT_RED_URL,
            'client_id': "client_2",
            'state': CLIENT_AUTHN_REQUEST['state'],
            'code': resp["code"],
            # TODO
            # 'code_verifier': 'ySfTlMpTEZPYU7H0XQZ75b3B568R5kkMkGRuRpQHOr1KNC9oimGnWygexLJuTyyT'
        }

        credentials = f"client_2:{CLIENT_1_PASSWD}"
        basic_auth = urlsafe_b64encode(credentials.encode("utf-8")).decode("utf-8")
        _basic_auth = f"Basic {basic_auth}"
        context.request_authorization = _basic_auth

        # TODO idpyoidc.server.exception.InvalidBranchID exception is caused cuz 'one!for!all' (user_id) is not found in Database.db (don't know how to put it there)
        token_resp = frontend.token_endpoint(context)
        _token_resp = json.loads(token_resp.message)
        assert _token_resp.get('error') == "invalid_request"


    def test_authorization_endpoint_refresh_without_consent(self, context, frontend):
        authn_req = self.get_authn_req(**{"prompt":"none"})
        client = self.prepare_call(context, frontend, authn_req)
        assert client['client_id'] == authn_req['client_id']
        res = frontend.authorization_endpoint(context)
        assert res.message == '{"error": "invalid_request", "error_description": "consent in prompt"}'


    def teardown_method(self):
        """ Clean up mongo """
        frontend = self.create_frontend(OIDCOP_CONF)
        if frontend.app.storage.client_db:
            frontend.app.storage.client_db.drop()
        if frontend.app.storage.session_db:
            frontend.app.storage.session_db.drop()



import json

from satosa.response import Response


class JsonResponse(Response):
    _content_type = "application/json"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if isinstance(self.message, list):
            self.message = self.message[0]

        if type(self.message) in (list, dict):
            self.message = json.dumps(self.message)



from idpyoidc.server.user_info import UserInfo


class UserInfoDict:
    def __init__(self, claims: dict):
        self.claims = claims

    def __getitem__(self, key):
        return self.claims


class SatosaOidcUserInfo(UserInfo):
    """Read only interface to a user info store"""

    def __init__(self, user_claims: dict = None):
        self.user_claims = user_claims
        self.load(user_claims)

    def load(self, user_claims):
        self.db = UserInfoDict(user_claims or {})

    def flush(self):
        self.db = UserInfoDict(claims={})



"""
The OpenID Connect frontend module for the satosa proxy
"""
import base64
import logging
import os
from urllib.parse import urlencode, urlparse, parse_qs
from datetime import datetime

from cryptojwt.key_jar import KeyJar
from idpyoidc.message.oauth2 import ResponseMessage
from idpyoidc.message.oidc import AccessTokenRequest
from idpyoidc.message.oidc import AuthnToken
from idpyoidc.message.oidc import AuthorizationErrorResponse
from idpyoidc.message.oidc import AuthorizationRequest
from idpyoidc.message.oidc import TokenErrorResponse
from idpyoidc.server.authn_event import create_authn_event
from idpyoidc.server.exception import ClientAuthenticationError
from idpyoidc.server.exception import NoSuchGrant
from idpyoidc.server.exception import InvalidClient
from idpyoidc.server.exception import UnAuthorizedClient
from idpyoidc.server.exception import UnknownClient
from idpyoidc.server.oidc.registration import random_client_id
from satosa.context import Context
try:
    from satosa.context import add_prompt_to_context
except ImportError:
    # TODO: remove after https://github.com/IdentityPython/SATOSA/pull/419 is merged
    def add_prompt_to_context(*args, **kwargs):
        pass
from satosa.frontends.base import FrontendModule
from satosa.internal import InternalData
import satosa.logging_util as lu
from satosa.response import SeeOther

from .core.application import oidcop_application as oidcop_app
from .core.claims import combine_claim_values
from .core.response import JsonResponse

IGNORED_HEADERS = ["cookie", "user-agent"]
logger = logging.getLogger(__name__)


class ExtendedContext(Context):  # pragma: no cover
    def __init__(self, **kwargs):
        super().__init__()
        self.http_headers = {}
        self.request_method = ""
        self.request_uri = ""
        self.request_authorization = ""


class OidcOpUtils(object):
    """
    Interoperability class between satosa and oidcop
    """

    def __init__(self):  # pragma: no cover
        self.app = None

    def get_client_info(self, client_id, context):
        _cinfo = context.cdb.get(client_id)
        if _cinfo:
            return _cinfo

        _cinfo = self.app.storage.get_client_by_id(client_id)
        if _cinfo:
            context.cdb = {client_id: _cinfo}

        return _cinfo

    def _load_cdb(self, context: ExtendedContext, client_id: str = None) -> dict:
        """
        gets client_id from local storage and updates the client DB
        """
        if client_id:
            client_id = client_id
        elif context.request and isinstance(context.request, dict):
            client_id = context.request.get("client_id")

        _ec = self.app.server.context

        if client_id:
            client = self.app.storage.get_client_by_id(client_id)
        elif "Basic " in getattr(context, "request_authorization", ""):
            # here even for introspection endpoint
            client = (
                    self.app.storage.get_client_by_basic_auth(
                        context.request_authorization)
                    or {}
            )
            client_id = client.get("client_id")
        elif context.request and context.request.get(
                "client_assertion"
        ):  # pragma: no cover
            # this is not a validation just a client detection
            # validation is demanded later to oidcop parse_request

            ####
            # WARNING: private_key_jwt can't be supported in SATOSA directly to token endpoint
            # because the user MUST always pass through the authorization endpoint
            ####
            token = AuthnToken().from_jwt(
                txt=context.request["client_assertion"],
                keyjar=KeyJar(),  # keyless keyjar
                verify=False,  # otherwise keyjar would contain the issuer key
            )
            client_id = token.get("iss")
            client = self.app.storage.get_client_by_id(client_id)
            
        elif "Bearer " in getattr(context, "request_authorization", ""):
            client = (
                    self.app.storage.get_client_by_bearer_token(
                        context.request_authorization)
                    or {}
            )
            client_id = client.get("client_id")
            
        else:  # pragma: no cover
            _ec.cdb = {}
            _msg = f"Client {client_id} not found!"
            logger.warning(_msg)
            raise InvalidClient(_msg)

        if client:
            _ec.cdb = {client_id: client}
            logger.debug(
                f"Loaded oidcop client from {self.app.storage}: {client}")
        else:  # pragma: no cover
            logger.info(f'Cannot find "{client_id}" in client DB')
            raise UnknownClient(client_id)

        # TODO - consider to handle also basic auth for clients ...
        # BUT specs are against!
        # https://openid.net/specs/openid-connect-registration-1_0.html#ReadRequest
        _rat = client.get("registration_access_token")
        if _rat:
            _ec.registration_access_token[_rat] = client["client_id"]
        else:
            _ec.registration_access_token = {}
        return client

    def _get_http_headers(self, context: ExtendedContext):
        """
        aligns parameters for oidcop interoperability needs
        """
        http_headers = {"headers": {}}

        # actually cookies are not used
        # if getattr(context, 'cookie', None):
        # for i in context.cookie.split(";"):
        # splitted = i.split("=")
        # if len(splitted) > 1:
        # _cookies.append(
        # {
        # "name": splitted[0].strip(),
        # "value": splitted[1].strip()
        # }
        # )
        # if _cookies:
        # http_headers["cookie"] = _cookies

        if getattr(context, "http_headers", None):
            http_headers = {
                "headers": {
                    k.lower(): v
                    for k, v in context.http_headers.items()
                    if k not in IGNORED_HEADERS
                },
                "method": context.request_method,
                "url": context.request_uri,
            }

        # for token and userinfo endpoint ... but also for authz endpoint if needed
        if getattr(context, "request_authorization", None):
            http_headers["headers"].update(
                {"authorization": context.request_authorization}
            )
        return http_headers

    def store_session_to_db(self, claims=None):
        sman = self.app.server.context.session_manager
        self.app.storage.store_session_to_db(sman, claims)
        logger.debug(f"Stored oidcop session to db: {sman.dump()}")

    def load_session_from_db(self, parse_req, http_headers):
        sman = self.app.server.context.session_manager
        claims = self.app.storage.load_session_from_db(
            parse_req, http_headers, sman)
        logger.debug(f"Loaded oidcop session from db: {sman.dump()}")
        
        return claims

    def _flush_endpoint_context_memory(self):
        """
        each OAuth2/OIDC request loads an oidcop session in memory
        this method will simply free the memory from any loaded session
        """
        _ec = self.app.server.context
        sman = _ec.session_manager
        sman.flush()

    def _load_session(self, parse_req, endpoint, http_headers):
        """
        actions to perform before an endpoint handles a new http request
        """
        self._flush_endpoint_context_memory()
        self.load_session_from_db(parse_req, http_headers)

    def _parse_request(
            self, endpoint, context: ExtendedContext, http_headers: dict = None
    ):
        """
        Returns a parsed OAuth2/OIDC request,
        used by Authorization, Token, Userinfo and Introspection enpoints views
        """
        http_headers = http_headers or self._get_http_headers(context)
        try:
            parse_req = endpoint.parse_request(
                context.request, http_info=http_headers,
            )
        except (
                InvalidClient,
                UnknownClient,
                UnAuthorizedClient,
                ClientAuthenticationError,
        ) as err:
            logger.error(err)
            response = JsonResponse(
                {"error": "unauthorized_client", "error_description": str(err)},
                status="403 Forbidden",
            )
            return self.send_response(response)
        return parse_req

    def _process_request(self, endpoint, context: Context, parse_req, http_headers):
        """
        Processes an OAuth2/OIDC request
        used by Authorization, Token, Userinfo and Introspection enpoints views
        """
        if isinstance(parse_req, JsonResponse):
            return self.send_response(parse_req)

        # do not handle prompt param by oidc-op, handle it here instead
        prompt_arg = parse_req.pop("prompt", None)
        if prompt_arg:
            add_prompt_to_context(context, " ".join(prompt_arg) if isinstance(prompt_arg, list) else prompt_arg)

        # save ACRs
        acr_values = parse_req.pop("acr_values", None)
        if acr_values:
            acr_values = acr_values if isinstance(acr_values, list) else acr_values.split(" ")
            context.decorate(Context.KEY_AUTHN_CONTEXT_CLASS_REF, acr_values)
            context.state[Context.KEY_AUTHN_CONTEXT_CLASS_REF] = acr_values

        try:
            proc_req = endpoint.process_request(
                parse_req, http_info=http_headers)
            return proc_req
        except Exception as err:  # pragma: no cover
            logger.error(
                f"In endpoint.process_request: {parse_req.__dict__} - {err}")
            response = JsonResponse(
                {
                    "error": "invalid_request",
                    "error_description": "request cannot be processed",
                },
                status="403 Forbidden",
            )
            return self.send_response(response)

    def _log_request(self, context: ExtendedContext, msg: str, level: str = "info"):
        _msg = f"{msg}: {context.request}"
        logline = lu.LOG_FMT.format(
            id=lu.get_session_id(context.state), message=_msg)
        getattr(logger, level)(logline)

    def handle_error(
            self, msg: str = None, excp: Exception = None, status: str = "403 Forbidden"
    ):  # pragma: no cover
        _msg = f'Something went wrong ... {excp or ""}'
        msg = msg or _msg
        logger.error(msg)
        response = JsonResponse(msg, status=status)
        return self.send_response(response)

    def send_response(self, response):
        self._flush_endpoint_context_memory()
        return response

    def dump_clients(self):  # pragma: no cover
        return self.app.server.context.cdb

    def dump_sessions(self):  # pragma: no cover
        return self.app.server.context.session_manager.dump()


class OidcOpEndpoints(OidcOpUtils):
    """Handles all the oidc endpoint"""

    def jwks_endpoint(self, context: Context):
        """
        Construct the JWKS document (served at /jwks).
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        return JsonResponse(self.jwks_public)

    def provider_info_endpoint(self, context: ExtendedContext):
        """
        Construct the provider configuration information
        served at /.well-known/openid-configuration.
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        endpoint = self.app.server.endpoint["provider_config"]
        logger.info(f'Request at the "{endpoint.name}" endpoint')
        http_headers = self._get_http_headers(context)

        parse_req = endpoint.parse_request(
            context.request, http_info=http_headers)
        proc_req = endpoint.process_request(parse_req, http_info=http_headers)

        info = endpoint.do_response(request=context.request, **proc_req)
        return JsonResponse(info["response"])

    def authorization_endpoint(self, context: ExtendedContext):
        """
        OAuth2 / OIDC Authorization endpoint
        Checks client_id and handles the authorization request
        """
        self._log_request(context, "Authorization endpoint request")
        self._load_cdb(context)

        endpoint = self.app.server.endpoint["authorization"]
        self._get_http_headers(context)
        internal_req = self._handle_authn_request(context, endpoint)
        if not isinstance(internal_req, InternalData):  # pragma: no cover
            return self.send_response(internal_req)

        return self.auth_req_callback_func(context, internal_req)

    def token_endpoint(self, context: ExtendedContext):
        """
        Handle token requests (served at /token).
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        self._log_request(context, "Token endpoint request")
        endpoint = self.app.server.endpoint["token"]
        http_headers = self._get_http_headers(context)

        try:
            self._load_cdb(context)
        except UnknownClient:
            return self.send_response(
                JsonResponse(
                    {
                        "error": "unauthorized_client",
                        "error_description": "unknown client",
                    }
                )
            )

        raw_request = AccessTokenRequest().from_urlencoded(urlencode(context.request))
        try:
            self._load_session(raw_request, endpoint, http_headers)
        except NoSuchGrant:
            _response = JsonResponse(
                {
                    "error": "invalid_request",
                    "error_description": "Not owner of token",
                },
                status="403 Forbidden",
            )
            return self.send_response(_response)

        # in token endpoint we cannot parse a request without having loaded cdb and session first
        parse_req = self._parse_request(endpoint, context, http_headers=http_headers)

        ec = endpoint.upstream_get("context")
        self._load_claims(ec)
        proc_req = self._process_request(endpoint, context, parse_req, http_headers)
        # flush as soon as possible, otherwise in case of an exception it would be
        # stored in the object ... until a next .load would happen ...
        ec.userinfo.flush()

        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return self.send_response(proc_req)
        elif isinstance(proc_req, TokenErrorResponse):
            return self.send_response(JsonResponse(proc_req.to_dict(), status="403 Forbidden"))

        # TODO: remove when migrate to idpy-oidc
        # PATCH https://github.com/UniversitaDellaCalabria/SATOSA-oidcop/issues/29
        if isinstance(proc_req["response_args"].get("scope", str), list):
            proc_req["response_args"]["scope"] = " ".join(
                proc_req["response_args"]["scope"]
            )
        # end PATCH

        # better return jwt or jwe here!
        self.store_session_to_db()
        response = JsonResponse(proc_req["response_args"])

        return self.send_response(response)

    def userinfo_endpoint(self, context: ExtendedContext):
        self._log_request(context, "Userinfo endpoint request")
        endpoint = self.app.server.endpoint["userinfo"]
        http_headers = self._get_http_headers(context)

        # everything depends on bearer access token here
        self._load_session({}, endpoint, http_headers)
        
        # not load the client from the session using the bearer token
        if self.dump_sessions():
            # load cdb from authz bearer token
            try:
                self._load_cdb(context)
            except Exception:
                logger.warning(
                    f"Userinfo endpoint request without any loadable client"
                )
                return self.send_response(
                    JsonResponse(
                        {"error": "invalid_client", "error_description": "<client not found>"},
                        status="403 Forbidden",
                    )
                )
        else:
            logger.warning(
                f"Userinfo endpoint request without any loadable sessions"
            )
            return self.send_response(
                JsonResponse(
                    {"error": "invalid_token", "error_description": "<no loadable session>"},
                    status="403 Forbidden",
                )
            )

        try:
            parse_req = self._parse_request(
                endpoint, context, http_headers=http_headers
            )
        except KeyError:
            return self.send_response(
                JsonResponse(
                    {"error": "invalid_token", "error_description": "<TOKEN>"},
                    status="403 Forbidden",
                )
            )

        ec = endpoint.upstream_get("context")
        self._load_claims(ec)
        proc_req = self._process_request(
            endpoint, context, parse_req, http_headers)
        # flush as soon as possible, otherwise in case of an exception it would be
        # stored in the object ... until a next .load would happen ...
        ec.userinfo.flush()

        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return self.send_response(proc_req)
        elif "error" in proc_req or "error" in proc_req.get("response_args", {}):
            return self.send_response(
                JsonResponse(
                    proc_req["response_args"]
                    if "response_args" in proc_req
                    else proc_req.to_dict(),
                    status="403 Forbidden",
                )
            )

        # better return jwt or jwe here!
        response = JsonResponse(proc_req["response_args"])

        self.store_session_to_db()
        return self.send_response(response)

    def _load_claims(self, endpoint_context):
        claims = {}
        sman = endpoint_context.session_manager
        for k, v in sman.dump()["db"].items():
            if v[0] == "idpyoidc.server.session.grant.Grant":
                sid = k
                claims = self.app.storage.get_claims_from_sid(sid)
                break
            else:  # pragma: no cover
                continue

        if not claims:
            logger.warning(
                "Can't find any suitable sid/claims from stored session"
            )

        # That's a patchy runtime definition of userinfo db configuration
        endpoint_context.userinfo.load(claims)

    def registration_read_endpoint(self, context: Context):
        """
        The Client Configuration Endpoint is an OAuth 2.0 Protected Resource
        that MAY be provisioned by the server for a specific Client to be able
        to view its registered information.

        The Client MUST use its Registration Access Token in all calls
        to this endpoint as an OAuth 2.0 Bearer Token
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        self._log_request(context, "Client Registration Read endpoint request")
        self._load_cdb(context)
        http_headers = self._get_http_headers(context)
        endpoint = self.app.server.endpoint["registration_read"]
        parse_req = self._parse_request(
            endpoint, context, http_headers=http_headers)
        proc_req = self._process_request(
            endpoint, context, parse_req, http_headers)
        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return self.send_response(proc_req)
        # better return jwt or jwe here!
        response = JsonResponse(proc_req["response_args"].to_dict())
        return self.send_response(response)

    def registration_endpoint(self, context: Context):
        """
        Handle the OIDC dynamic client registration.
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        self._log_request(context, "Client Registration endpoint request")
        http_headers = self._get_http_headers(context)
        endpoint = self.app.server.endpoint["registration"]
        parse_req = self._parse_request(
            endpoint, context, http_headers=http_headers)
        proc_req = self._process_request(
            endpoint, context, parse_req, http_headers)
        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return self.send_response(proc_req)
        # store client to storage
        client_data = context.request
        client_data["client_id"] = random_client_id(
            reserved=self.app.storage.get_registered_clients_id()
        )
        self.app.storage.insert_client(client_data)
        return JsonResponse(client_data)

    def introspection_endpoint(self, context: Context):
        self._log_request(context, "Token Introspection endpoint request")
        endpoint = self.app.server.endpoint["introspection"]
        http_headers = self._get_http_headers(context)

        self._load_cdb(context)
        self._load_session(context.request, endpoint, http_headers)
        parse_req = self._parse_request(
            endpoint, context, http_headers=http_headers)
        proc_req = self._process_request(
            endpoint, context, parse_req, http_headers)
        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return self.send_response(proc_req)

        # better return jwt or jwe here!
        response = JsonResponse(proc_req["response_args"].to_dict())

        self._flush_endpoint_context_memory()
        return self.send_response(response)


class OidcOpFrontend(FrontendModule, OidcOpEndpoints):
    """
    OpenID Connect frontend module based on idpy oidcop
    """

    def __init__(
            self, auth_req_callback_func, internal_attributes, conf, base_url, name
    ):
        super().__init__(auth_req_callback_func, internal_attributes, base_url, name)
        self.app = oidcop_app(conf)
        # Why not
        # self.config = self.app.server.conf
        self.config = self.app.srv_config
        jwks_public_path = self.config["key_conf"]["public_path"]
        with open(jwks_public_path) as f:
            self.jwks_public = f.read()

        # registered endpoints will be filled by self.register_endpoints
        self.endpoints = None

    def register_endpoints(self, backend_names):
        """
        See super class satosa.frontends.base.FrontendModule
        :type backend_names: list[str]
        :rtype: list[(str, ((satosa.context.Context, Any) -> satosa.response.Response, Any))]
        :raise ValueError: if more than one backend is configured
        """
        url_map = [
            (v["path"], getattr(self, f"{k}_endpoint"))
            for k, v in self.config.endpoint.items()
        ]

        # add jwks.json webpath
        uri_path = self.config["key_conf"]["uri_path"]
        url_map.append((uri_path, self.jwks_endpoint))

        logger.debug(f"Loaded OIDC Provider endpoints: {url_map}")
        self.endpoints = url_map
        return url_map

    def _handle_authn_request(self, context: ExtendedContext, endpoint):
        """
        Parse and verify the authentication request into an internal request.
        :type context: satosa.context.Context
        :rtype: satosa.internal.InternalData

        :param context: the current context
        :return: the internal request
        """
        self._log_request(context, "OIDC Authorization request from client")

        http_headers = self._get_http_headers(context)
        self._load_cdb(context)
        parse_req = self._parse_request(
            endpoint, context, http_headers=http_headers)
        if isinstance(parse_req, AuthorizationErrorResponse):
            logger.debug(f"{context.request}, {parse_req._dict}")
            return self.send_response(JsonResponse(parse_req._dict))

        self._load_session(parse_req, endpoint, http_headers)
        proc_req = self._process_request(
            endpoint, context, parse_req, http_headers)
        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return proc_req

        try:
            endpoint.do_response(request=context.request, **proc_req)
        except Exception as excp:  # pragma: no cover
            # TODO - something to be done with the help of unit test
            # this should be for humans if auth code flow
            # and JsonResponse for other flows ...
            self.handle_error(excp=excp)

        context.state[self.name] = {"oidc_request": context.request}

        client_id = parse_req.get("client_id")
        _client_conf = endpoint.upstream_get("context").cdb[client_id]
        client_name = _client_conf.get("client_name")
        subject_type = _client_conf.get("subject_type", "pairwise")
        if client_name:
            requester_name = [{"lang": "en", "text": client_name}]
        else:  # pragma: no cover
            requester_name = None

        internal_req = InternalData(
            subject_type=subject_type,
            requester=client_id,
            requester_name=requester_name,
        )

        _claims_supported = self.config["capabilities"]["claims_supported"]

        # TODO - additional filter here?
        # _approved_attributes = self._get_approved_attributes(
        # _claims_supported, authn_req
        # )
        internal_req.attributes = self.converter.to_internal_filter(
            "openid", _claims_supported
        )

        # TODO - have a default backend, otherwise exception here ...
        context.target_backend = self.app.default_target_backend
        context.internal_data = internal_req
        return internal_req

    def handle_authn_request(self, context: ExtendedContext):
        """
        Handle an authentication request and pass it on to the backend.
        :type context: satosa.context.Context
        :rtype: satosa.response.SeeOther

        :param context: the current context
        :return: HTTP response to the client
        """
        endpoint = self.app.server.endpoint["authorization"]
        internal_req = self._handle_authn_request(context, endpoint)
        if not isinstance(internal_req, InternalData):
            return self.send_response(internal_req)
        return self.auth_req_callback_func(context, internal_req)

    def _handle_backend_response(self, context: ExtendedContext, internal_resp):
        """
        Called by handle_authn_response, once a backend made its work
        :type context: satosa.context.Context
        :rtype: satosa.response.Response

        :param context: the current context
        :param internal_resp: satosa internal data
        :type internal_resp: satosa.internal.InternalData
        :return: HTTP response to the client
        """
        http_headers = self._get_http_headers(context)
        oidc_req = context.state[self.name]["oidc_request"]
        endpoint = self.app.server.endpoint["authorization"]
        self._load_cdb(context, client_id=oidc_req["client_id"])

        # not using self._parse_request cause of "Missing required attribute 'response_type'"
        parse_req = AuthorizationRequest().from_urlencoded(urlencode(oidc_req))
        # not really needed here ... because nothing have been dumped before
        # self._load_session(parse_req, endpoint, http_headers)
        proc_req = self._process_request(
            endpoint, context, parse_req, http_headers)

        if isinstance(proc_req, JsonResponse):  # pragma: no cover
            return self.send_response(proc_req)

        client_id = parse_req["client_id"]
        sub = internal_resp.subject_id

        authn_event = create_authn_event(
            uid=sub,
            salt=base64.b64encode(os.urandom(self.app.salt_size)).decode(),
            authn_info=internal_resp.auth_info.auth_class_ref,
            # TODO: authn_time=datetime.fromisoformat(internal_resp.auth_info.timestamp).timestamp(),
        )

        _ec = endpoint.upstream_get("context")
        _token_usage_rules = _ec.authn_broker.get_method_by_id("user")

        session_manager = _ec.session_manager
        client = self.app.storage.get_client_by_id(client_id)
        client_subject_type = client.get("subject_type", "public")
        _session_id = session_manager.create_session(
            authn_event=authn_event,
            auth_req=parse_req,
            user_id=sub,
            client_id=client_id,
            sub_type=client_subject_type,
            token_usage_rules=_token_usage_rules,
        )

        try:
            # _args is a dict that contains:
            #  - idpyoidc.message.oidc.AuthorizationResponse
            #  - session_id
            #  - cookie (only need for logout -> not yet supported by Satosa)
            _args = endpoint.authz_part2(
                user=sub,
                session_id=_session_id,
                request=parse_req,
                authn_event=authn_event,
            )
        except ValueError as excp:  # pragma: no cover
            # TODO - cover with unit test and add some satosa logging ...
            return self.handle_error(excp=excp)
        except Exception as excp:  # pragma: no cover
            return self.handle_error(excp=excp)

        if isinstance(_args, ResponseMessage) and "error" in _args:  # pragma: no cover
            return self.send_response(JsonResponse(_args, status="403 Forbidden"))
        elif isinstance(
                _args.get("response_args"), AuthorizationErrorResponse
        ):  # pragma: no cover
            rargs = _args.get("response_args")
            logger.error(rargs)
            response = JsonResponse(rargs.to_json(), status="403 Forbidden")
            return self.send_response(response)

        info = endpoint.do_response(request=parse_req, **proc_req)
        info_response = info["response"]
        _response_placement = info.get(
            "response_placement", endpoint.response_placement
        )
        if _response_placement == "url":
            data = _args["response_args"].to_dict()
            url_components = urlparse(info_response)
            original_params = parse_qs(url_components.query)
            merged_params = {**original_params, **data}
            updated_query = urlencode(merged_params, doseq=True)
            redirect_url = url_components._replace(query=updated_query).geturl()
            logger.debug(f"Redirect to: {redirect_url}")
            resp = SeeOther(redirect_url)
        else:  # pragma: no cover
            self._flush_endpoint_context_memory()
            raise NotImplementedError()

        # I don't flush in-mem db because it will be flushed by handle_authn_response
        return resp

    def handle_authn_response(self, context: ExtendedContext, internal_resp):
        """
        See super class method satosa.frontends.base.FrontendModule#handle_authn_response
        :type context: satosa.context.Context
        :type internal_resp: satosa.internal.InternalData
        :rtype satosa.response.SeeOther
        """
        _claims = self.converter.from_internal(
            "openid", internal_resp.attributes)
        claims = {k: v for k, v in _claims.items() if v}
        combined_claims = dict(
            [i for i in combine_claim_values(claims.items())])

        response = self._handle_backend_response(context, internal_resp)
        # TODO - why should we have to delete it?
        del context.state[self.name]

        # store oidc session with user claims
        self.store_session_to_db(claims=combined_claims)
        return self.send_response(response)

    def handle_backend_error(self, exception):
        """
        See super class satosa.frontends.base.FrontendModule
        :type exception: satosa.exception.SATOSAError
        :rtype: oic.utils.http_util.Response
        """
        auth_req = AuthorizationRequest().from_urlencoded(
            urlencode(exception.state[self.name]["oidc_request"])
        )
        msg = exception.message
        error_resp = AuthorizationErrorResponse(
            error="access_denied",
            error_description=msg,
            # If the client sent us a state parameter, we should reflect it back according to the
            # spec
            **({"state": auth_req["state"]} if "state" in auth_req else {}),
        )
        logline = lu.LOG_FMT.format(
            id=lu.get_session_id(exception.state), message=msg)
        logger.info(logline)
        return SeeOther(
            error_resp.request(
                auth_req["redirect_uri"], auth_req["response_type"] != ["code"]
            )
        )



import base64
import copy
import datetime
import json
import logging
import hmac

from idpyoidc.server.exception import NoSuchGrant
import pymongo

from .base import SatosaOidcStorage
from idpyoidc.server.session.manager import SessionManager

logger = logging.getLogger(__name__)


class Mongodb(SatosaOidcStorage):
    session_attr_map = {
        "idpyoidc.server.session.info.UserSessionInfo": "sub",
        "idpyoidc.server.session.info.ClientSessionInfo": "client_id",
        "idpyoidc.server.session.grant.Grant": "grant_id",
    }
    token_attr_map = {
        "idpyoidc.server.session.token.AuthorizationCode": "authorization_code",
        "idpyoidc.server.session.token.AccessToken": "access_token",
        "idpyoidc.server.session.token.RefreshToken": "refresh_token",
        "idpyoidc.server.session.token.IDToken": "id_token",
    }

    def __init__(self, storage_conf: dict, url: str, connection_params: dict = None):
        self.storage_conf = storage_conf
        self.url = url
        self.connection_params = connection_params

        # this must be fork safe :)
        self.client = None
        self.db = None
        self.client_db = None
        self.session_db = None

    def _connect(self):
        if not self.client or not self.client.server_info():
            self.client = pymongo.MongoClient(
                self.url, **self.connection_params)
            self.db = getattr(self.client, self.storage_conf["db_name"])
            self.client_db = self.db[self.storage_conf["collections"]["client"]]
            self.session_db = self.db[self.storage_conf["collections"]["session"]]

    def get_client_by_id(self, client_id: str):
        self._connect()
        return self.client_db.find_one({"client_id": client_id}) or {}

    def store_session_to_db(self, session_manager: SessionManager, claims: dict):
        ses_man_dump = session_manager.dump()
        _db = ses_man_dump["db"]
        data = {
            "expires_at": 0,
            "sub": "",
            "client_id": "",
            "grant_id": "",
            "sid": "",
            "sid_encrypted": "",
            "authorization_code": "",
            "access_token": "",
            "id_token": "",
            "refresh_token": "",
            "claims": claims or {},
            "dump": json.dumps(_db),
            "key": ses_man_dump['crypt_config']['kwargs']["password"],
            "salt": ses_man_dump['crypt_config']['kwargs']["salt"]
        }

        for k, v in _db.items():
            # TODO: ask to roland to have something better than this
            if len(k) > 128 and ";;" not in k and v[0] == "idpyoidc.server.session.grant.Grant":
                data["sid_encrypted"] = k
                continue
    
            classname = v[0]
            field_name = self.session_attr_map[classname]
            if field_name == "sub":
                data["client_id"] = list(_db.keys())[1].split(";")[-1]
                data[field_name] = _db[list(_db.keys())[2]][1]['sub']
            elif field_name == "client_id":
                data["grant_id"] = v[1]["subordinate"][0]
            elif field_name == "grant_id":
                _exp_time = datetime.datetime.fromtimestamp(v[1]["expires_at"])
                data["expires_at"] = _exp_time
                data["revoked"] = v[1]["revoked"]
                # data['sub'] = v[1]['sub']
                data["sid"] = k

                iss_tokens = {}
                for i in v[1]["issued_token"]:
                    k = list(i.keys())[0]
                    iss_tokens[k] = i[k]["value"]

                for token_type, attr in self.token_attr_map.items():
                    data[attr] = iss_tokens.get(token_type)

        logger.debug(f"Stored oidcop session data to MongoDB: {data}")

        self._connect()
        q = {"grant_id": data["grant_id"]}
        grant = self.session_db.find_one(q)
        if grant:
            # if update preserve the claims
            data["claims"] = grant["claims"]
            self.session_db.update_one(q, {"$set": data})
        else:
            self.session_db.insert_one(data)

    def load_session_from_db(
        self, parse_req, http_headers: dict, session_manager: SessionManager, **kwargs
    ) -> dict:
        """
        This method detects some usefull elements for doing a lookup in the session storage
        then loads the session in-memory

        It doesn't want to do any validation but only loading a session inmemory
        Security validation will be made later by oidcop in process_request
        """
        data = {}
        _q = {}
        http_authz = http_headers.get("headers", {}).get("authorization", "")

        if parse_req.get("grant_type") == "authorization_code":
            # here for auth code flow and token endpoint only
            _q = {
                "authorization_code": parse_req["code"],
                "client_id": parse_req.get("client_id")
                or self.get_client_id_by_basic_auth(http_authz),
            }
        elif http_authz and "Basic " not in http_authz:
            # here for userinfo endpoint
            # exclude Basic auth: we want only bearer and dpop here!
            _q = {
                "access_token": http_authz.replace("Bearer ", ""),
            }
        elif parse_req.get("token"):
            _q = {
                "access_token": parse_req["token"],
            }
        elif parse_req.get("grant_type") == "refresh_token":
            _q = {
                "refresh_token": parse_req["refresh_token"],
            }
        else:
            logger.warning(
                f"load_session_from_db can't find any active session from: {parse_req}"
            )
            return data

        self._connect()
        res = self.session_db.find_one(_q)
        if res:
            data["key"] = res["key"]
            data["salt"] = res["salt"]
            data["db"] = json.loads(res["dump"])
            session_manager.flush()
            session_manager.load(data)
        elif 'client_id' in _q:
            raise NoSuchGrant('The client has not been issued the grant')
        return data

    def get_claims_from_sid(self, sid: str):
        self._connect()
        res = self.session_db.find_one({"sid": sid})
        if res:
            return res["claims"]

    def insert_client(self, client_data: dict):
        _client_data = copy.deepcopy(client_data)
        self._connect()
        client_id = _client_data["client_id"]
        if self.get_client_by_id(client_id):
            logger.debug(
                f"OIDC Client {client_id} already present in the client db")
            return
        self.client_db.insert_one(_client_data)

    def get_client_creds_from_basic_auth(self, request_authorization: str):
        cred = base64.b64decode(
            request_authorization.replace("Basic ", "").encode())
        if not cred:
            return

        cred = cred.decode().split(":")
        if len(cred) == 2:
            return cred

    def get_client_id_by_basic_auth(self, request_authorization: str):
        cred = self.get_client_creds_from_basic_auth(request_authorization)

        if len(cred) == 2:
            client_id = cred[0]
            return client_id

    def get_client_by_basic_auth(self, request_authorization: str):
        cred = self.get_client_creds_from_basic_auth(request_authorization)
        
        if len(cred) == 2:
            client_id = cred[0]
            client_secret = cred[1]

            self._connect()
            client = self.client_db.find_one(
                {"client_id": client_id}
            )
            return client if "client_secret" in client and hmac.compare_digest(client["client_secret"], client_secret) else None

    def get_client_by_bearer_token(self, request_authorization: str):
        access_token = request_authorization.replace("Bearer ", "")
        return self.session_db.find_one({"access_token": access_token})

    def get_registered_clients_id(self):
        self._connect()
        res = self.client_db.distinct("client_id")
        return res



from idpyoidc.server.user_authn.user import UserAuthnMethod


class SatosaAuthnMethod(UserAuthnMethod):
    """
    Dummy approach to get SATOSA working with oidcop as it is
    """

    def __call__(self, **kwargs):
        """
        Display user interaction.

        :param args:
        :param kwargs:
        :return:
        """
        # raise NotImplementedError

    def verify(self, *args, **kwargs):
        """
        Callback to verify user input
        :return: username of the authenticated user
        """
        # raise NotImplementedError



class SatosaOidcStorage(object):
    def get_client_by_id(self, client_id: str, expired: bool = True):
        raise NotImplementedError()

    def store_session_to_db(self, session_manager, **kwargs):
        raise NotImplementedError()

    def load_session_from_db(self, req_args, http_headers, session_manager, **kwargs):
        raise NotImplementedError()

    def get_claims_from_sid(self, sid: str):
        raise NotImplementedError()

    def insert_client(self, client_data: dict):
        raise NotImplementedError()

    def get_client_by_basic_auth(self, request_authorization: str):
        raise NotImplementedError()

    def get_registered_clients_id(self):
        raise NotImplementedError()



from glob import glob
from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()


_pkg_name = "satosa_oidcop"

setup(
    name=_pkg_name,
    description="SATOSA Frontend based on idetity python oidcop",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    url="https://github.com/UniversitaDellaCalabria/satosa-oidcop",
    author="Giuseppe De Marco",
    author_email="giuseppe.demarco@unical.it",
    license="License :: OSI Approved :: GNU Affero General Public License v3",
    packages=[f"{_pkg_name}"],
    package_dir={f"{_pkg_name}": f"{_pkg_name}"},
    package_data={
        f"{_pkg_name}": [
            i.replace(f"{_pkg_name}/", "")
            for i in glob(f"{_pkg_name}/**", recursive=True)
        ]
    },
    install_requires=[
        "satosa>=8.0.0",
        "pymongo>=3.11,<5.0",
        "idpyoidc>=4.1.0,<4.2.0",
    ],
)
